/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef HAVE_LADCCA
#include <ladcca/ladcca.h>
extern cca_client_t * cca_client;
#endif

#include <gtk/gtk.h>
#include <ladspa.h>

#include "gui.h"
#include "xmalloc.h"
#include "lock-free-fifo.h"
#include "callbacks.h"
#include "control-message.h"

void add_callback (GtkMenuItem * menuitem, gpointer user_data) {
  guiinfo_t * gui_info;
  plugin_t * plugin;
  ladspa_chain_link_t * chain_link;
  int err;
  ctrlmsg_t ctrlmsg;
  
  gui_info = (guiinfo_t *) user_data;
  
  /* get the selected plugin out of the gobject data thing */
  plugin = g_object_get_data (G_OBJECT(menuitem), "jack-rack-plugin");
  if (!plugin) {
    printf ("%s: no plugin!\n", __FUNCTION__);
  }
  
  /* check whether or not the plugin is RT capable and confirm with the user if it isn't */
  if (!LADSPA_IS_HARD_RT_CAPABLE(plugin->descriptor->Properties)) {
    GtkWidget * dialog;
    gint response;

    dialog = gtk_message_dialog_new (GTK_WINDOW(gui_info->main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT|GTK_DIALOG_MODAL,
               GTK_MESSAGE_WARNING,
               GTK_BUTTONS_YES_NO,
               "Plugin not RT capable\n\nThe plugin '%s' does not describe itself as being capable of real-time operation.  You may experience drop outs or jack may even kick us out if you use it.\n\nAre you sure you want to add this plugin?",
               plugin->name);
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    
    if (response != GTK_RESPONSE_YES) {
      return;
    }
  }

  /* create the chain link */
  chain_link = xmalloc (sizeof (ladspa_chain_link_t));

  err = init_chain_link (chain_link, plugin);
  if (err) {
    GtkWidget * dialog;
    
    dialog = gtk_message_dialog_new (GTK_WINDOW(gui_info->main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT,
               GTK_MESSAGE_ERROR,
               GTK_BUTTONS_CLOSE,
               "Error loading file plugin '%s' from file '%s'",
               plugin->name, plugin->object_file->filename);
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    free (chain_link);
    return;
  }
  
  /* send the chain link off to the process() callback */
  ctrlmsg.type = CTRLMSG_ADD;
  ctrlmsg.pointer = chain_link;
  ctrlmsg.second_pointer = plugin;
  lff_write (gui_info->gui_to_process, &ctrlmsg);

}

void load_callback (GtkButton * button, gpointer user_data) {}
void save_callback (GtkButton * button, gpointer user_data) {}

void quit_callback (GtkButton * button, gpointer user_data) {
  guiinfo_t * gui_info;
  ctrlmsg_t ctrlmsg;
  
  gui_info = user_data;
  
  jack_deactivate (gui_info->process_info->jack_client);
  
  ctrlmsg.type = CTRLMSG_QUIT;
  lff_write (gui_info->gui_to_process, &ctrlmsg);
  
  gtk_main_quit ();
}

void slot_change_callback (GtkMenuItem * menuitem, gpointer user_data) {
  ctrlmsg_t ctrlmsg;
  plugin_slot_t * plugin_slot;
  ladspa_chain_link_t * chain_link;
  plugin_t * plugin;
  int err;
  
  plugin_slot = user_data;
  
  if (global_gui_info.moving_slot) return;
  
  plugin = g_object_get_data (G_OBJECT(menuitem), "jack-rack-plugin");

  /* check whether or not the plugin is RT capable and confirm with the user if it isn't */
  if (!LADSPA_IS_HARD_RT_CAPABLE(plugin->descriptor->Properties)) {
    GtkWidget * dialog;
    gint response;

    dialog = gtk_message_dialog_new (GTK_WINDOW(global_gui_info.main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT|GTK_DIALOG_MODAL,
               GTK_MESSAGE_WARNING,
               GTK_BUTTONS_YES_NO,
               "Plugin not RT capable\n\nThe plugin '%s' does not describe itself as being capable of real-time operation.  You may experience drop outs or jack may even kick us out if you use it.\n\nAre you sure you want change to this plugin?",
               plugin->name);
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    
    if (response != GTK_RESPONSE_YES) {
      return;
    }
  }
  
  chain_link = xmalloc (sizeof (ladspa_chain_link_t));

  err = init_chain_link (chain_link, plugin);
  if (err) {
    GtkWidget * dialog;
    
    dialog = gtk_message_dialog_new (GTK_WINDOW(global_gui_info.main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT,
               GTK_MESSAGE_ERROR,
               GTK_BUTTONS_CLOSE,
               "Error loading file plugin '%s' from file '%s'",
               plugin->name, plugin->object_file->filename);
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    free (chain_link);
    return;
  }

  ctrlmsg.type = CTRLMSG_CHANGE;
  ctrlmsg.number = plugin_slot->index;
  ctrlmsg.pointer = chain_link;
  ctrlmsg.second_pointer = chain_link;
  
  lff_write (global_gui_info.gui_to_process, &ctrlmsg);
}

void slot_move_callback (GtkButton * button, gpointer user_data) {
  ctrlmsg_t ctrlmsg;
  plugin_slot_t * plugin_slot;
  gint up;

  plugin_slot = user_data;
  up = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(button), "jack-rack-up"));
  
  /* this is because we uses indicies */
  if (global_gui_info.moving_slot) {
    return;
  }
  
  /* check the logic of what we're doing */
  if ((up && !plugin_slot->prev) ||
      (!up && !plugin_slot->next)) {
    return;
  }
  
  global_gui_info.moving_slot = TRUE;
  
  
  ctrlmsg.type = CTRLMSG_MOVE;
  ctrlmsg.number = plugin_slot->index;
  ctrlmsg.pointer = GINT_TO_POINTER (up);
  ctrlmsg.second_pointer = plugin_slot;
  
  lff_write (global_gui_info.gui_to_process, &ctrlmsg);
  
}

void slot_remove_callback (GtkButton * button, gpointer user_data) {
  ctrlmsg_t ctrlmsg;
  plugin_slot_t * plugin_slot;
  
  plugin_slot = (plugin_slot_t *) user_data;
  
  ctrlmsg.type = CTRLMSG_REMOVE;
  ctrlmsg.number = plugin_slot->index;
  ctrlmsg.second_pointer = plugin_slot;
  
  lff_write (global_gui_info.gui_to_process, &ctrlmsg);
}


gboolean slot_ablise_callback (GtkWidget * button, GdkEventButton *event, gpointer user_data) {
  if (event->type == GDK_BUTTON_PRESS) {
    plugin_slot_t * plugin_slot; 
    ctrlmsg_t ctrlmsg;
    gint ablise;
  
    plugin_slot = (plugin_slot_t *) user_data;
    ablise = (plugin_slot->enabled ? FALSE : TRUE);
  
    ctrlmsg.type = CTRLMSG_ABLE;
    ctrlmsg.number = plugin_slot->index;
    ctrlmsg.pointer = GINT_TO_POINTER(ablise);
    ctrlmsg.second_pointer = plugin_slot;
  
    lff_write (global_gui_info.gui_to_process, &ctrlmsg);
  
    return TRUE;
  }
  
  return FALSE;
}

void slot_lock_all_callback (GtkToggleButton * button, gpointer user_data) {
  gboolean on;
  plugin_slot_t * plugin_slot;
  unsigned long i;
  
  on = gtk_toggle_button_get_active (button);
  
  plugin_slot = (plugin_slot_t *) user_data;
  
  plugin_slot->plugin_settings[plugin_slot->plugin->collection_index].lock_all = on;

  for (i = 0; i < plugin_slot->plugin->control_port_count; i++) {
  
    gtk_widget_set_sensitive (
      plugin_slot->port_control_holders[i].lock,
      on ? FALSE : TRUE);
      
    if (on) {
      plugin_slot->port_control_holders[i].locked = TRUE;
    } else {
      plugin_slot->port_control_holders[i].locked =
        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(plugin_slot->port_control_holders[i].lock));
    }
    
  }
  
}

void control_lock_callback (GtkToggleButton * button, gpointer user_data) {
  port_controls_holder_t * port_controls_holder;
  
  port_controls_holder = (port_controls_holder_t *) user_data;
  
  port_controls_holder->locked = gtk_toggle_button_get_active (button);

  port_controls_holder->plugin_slot->
    plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
    locks[port_controls_holder->port_index] = port_controls_holder->locked;
}

void set_widget_text_colour (GtkWidget * widget,
                         const char * colour_str) {
  GdkColor colour;
  GdkColormap * colourmap;
  
  colourmap = gtk_widget_get_colormap (widget);
  
  gdk_color_parse (colour_str, &colour);
  gdk_colormap_alloc_color (colourmap, &colour, FALSE, TRUE);
  gtk_widget_modify_text (widget, GTK_STATE_NORMAL, &colour);
}

void control_float_callback (GtkRange * range, gpointer user_data) {
  port_controls_holder_t * port_controls_holder;
  gchar * str;
  LADSPA_Data value;
  int channel = 0;
  
  port_controls_holder = (port_controls_holder_t *) user_data;
  
  value = gtk_range_get_value (range);
  
  str = g_strdup_printf ("%f", value);
  
  if (port_controls_holder->mono) {

    /* get which channel we're using from the g object data stuff */
    channel = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(range), "jack-rack-mono-channel"));
    
    /* write to the fifo */
    lff_write (port_controls_holder->controls.mono[channel].control_fifo, &value);
    
    /* print the value in the text box */
    gtk_entry_set_text (GTK_ENTRY(port_controls_holder->controls.mono[channel].text), str);
    
    /* store the value */
    port_controls_holder->plugin_slot->
      plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
      control_values[channel][port_controls_holder->port_index] = value;
              
    
    /* possibly set our pair */
    if (port_controls_holder->locked) {
      gtk_range_set_value (GTK_RANGE(port_controls_holder->controls.mono[1-channel].control),
                           gtk_range_get_value (range));
    }
    
    
    set_widget_text_colour (port_controls_holder->controls.mono[channel].text, "Black");
  } else {
    lff_write (port_controls_holder->controls.stereo.control_fifo, &value);

    /* print the value in the text box */
    gtk_entry_set_text (GTK_ENTRY(port_controls_holder->controls.stereo.text), str);

    /* store the value */
    port_controls_holder->plugin_slot->
      plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
      control_values[0][port_controls_holder->port_index] = value;
    
    set_widget_text_colour (port_controls_holder->controls.stereo.text, "Black");
  }
  
  g_free (str);
  
}

void control_float_text_callback (GtkEntry * entry, gpointer user_data) {
  port_controls_holder_t * port_controls_holder;
  LADSPA_Data value;
  const char * str;
  char * endptr;
  GtkAdjustment * adjustment;
  int channel = 0;
  
  port_controls_holder = (port_controls_holder_t *) user_data;
  
  str = gtk_entry_get_text (entry);
  value = strtof (str, &endptr);
  
  /* check for errors */
  if (port_controls_holder->mono) {
    adjustment = gtk_range_get_adjustment (
                 GTK_RANGE (port_controls_holder->controls.mono[0].control));
  } else {
    adjustment = gtk_range_get_adjustment (
                 GTK_RANGE (port_controls_holder->controls.stereo.control));
  }
  
  if (value == HUGE_VALF ||
      value == -HUGE_VALF ||
      endptr == str ||
      value < adjustment->lower ||
      value > adjustment->upper) {
    /* set the entry's text colour */
    set_widget_text_colour (GTK_WIDGET(entry), "DarkRed");
    return;
  }
  
  
  if (port_controls_holder->mono) {
    /* get which channel we're using from the g object data stuff */
    channel = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(entry), "jack-rack-mono-channel"));
    
    
    
    /* set the value in range */
    gtk_range_set_value (GTK_RANGE(port_controls_holder->controls.mono[channel].control), value);
    
    /* possibly set our pair */
    if (port_controls_holder->locked) {
      gtk_range_set_value (GTK_RANGE(port_controls_holder->controls.mono[1-channel].control), value);
    }
    
  } else {
    /* set the value in range */
    gtk_range_set_value (GTK_RANGE(port_controls_holder->controls.stereo.control), value);
  }
  
  set_widget_text_colour (GTK_WIDGET(entry), "Black");
}


void control_bool_callback (GtkToggleButton * button, gpointer user_data) {
  port_controls_holder_t * port_controls_holder;
  gboolean on;
  LADSPA_Data data;
  
  port_controls_holder = (port_controls_holder_t *) user_data;
  
  on = gtk_toggle_button_get_active (button);
  data = on ? 1.0 : -1.0;
  
  if (port_controls_holder->mono) {
    int channel;

    /* get which channel we're using from the g object data stuff */
    channel = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(button), "jack-rack-mono-channel"));
    
    /* write to the fifo */
    lff_write (port_controls_holder->controls.mono[channel].control_fifo, &data);
    
    /* store value */
    port_controls_holder->plugin_slot->
      plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
      control_values[channel][port_controls_holder->port_index] = data;
      
    /* possibly set other control */
    if (port_controls_holder->locked) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON
                                    (port_controls_holder->controls.mono[1-channel].control),
                                    on);
    }
    
  } else {
    lff_write (port_controls_holder->controls.stereo.control_fifo, &data);
    
    /* store value */
    port_controls_holder->plugin_slot->
      plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
      control_values[0][port_controls_holder->port_index] = data;
  }
}
void control_int_callback (GtkSpinButton * spinbutton, gpointer user_data) {
  port_controls_holder_t * port_controls_holder;
  int value;
  LADSPA_Data data;
  
  port_controls_holder = (port_controls_holder_t *) user_data;
  
  value = gtk_spin_button_get_value_as_int (spinbutton);
  data = value;
  
  if (port_controls_holder->mono) {
    int channel;

    /* get which channel we're using from the g object data stuff */
    channel = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(spinbutton), "jack-rack-mono-channel"));
    
    /* write to the fifo */
    lff_write (port_controls_holder->controls.mono[channel].control_fifo, &data);
    
    /* store value */
    port_controls_holder->plugin_slot->
      plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
      control_values[channel][port_controls_holder->port_index] = data;
      
    /* possibly set other control */
    if (port_controls_holder->locked) {
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (port_controls_holder->controls.mono[1-channel].control),
                                 value);
    }
    
  } else {
    lff_write (port_controls_holder->controls.stereo.control_fifo, &data);
    
    /* store value */
    port_controls_holder->plugin_slot->
      plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
      control_values[0][port_controls_holder->port_index] = data;
  }
}


/** callback for plugin menu buttons (Add and the plugin slots' change buttons) */
gint plugin_button_callback (GtkWidget *widget, GdkEvent *event) {

  if (event->type == GDK_BUTTON_PRESS) {
    GdkEventButton * bevent = (GdkEventButton *) event;
    gtk_menu_popup (GTK_MENU (widget), NULL, NULL, NULL, NULL,
                    bevent->button, bevent->time);
    /* Tell calling code that we have handled this event; the buck
     * stops here. */
    return TRUE;
  }
  
  /* Tell calling code that we have not handled this event; pass it on. */
  return FALSE;
}

/* do the process->gui message processing */
gboolean idle_callback (gpointer data) {
  static gboolean shown_shutdown_msg = FALSE;
  ctrlmsg_t ctrlmsg;
  guiinfo_t * gui_info;
  ladspa_chain_link_t * chain_link;
  plugin_t * plugin;
  plugin_slot_t * plugin_slot;
  int enabled;
  
  gui_info = (guiinfo_t *) data;
  
  /* jack shut us down */
  if (gui_info->shutdown && !shown_shutdown_msg) {
    GtkWidget * dialog;
    gint response;

    dialog = gtk_message_dialog_new (GTK_WINDOW(gui_info->main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT|GTK_DIALOG_MODAL,
               GTK_MESSAGE_ERROR,
               GTK_BUTTONS_CLOSE,
               "JACK client thread shut down by server\n\nJACK, the bastard, kicked us out.  You'll have to restart I'm afraid.  Sorry.");
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    
    
    
    gtk_widget_set_sensitive (gui_info->plugin_box, FALSE);
    gtk_widget_set_sensitive (gui_info->add, FALSE);
    shown_shutdown_msg = TRUE;
  }
  
  while (lff_read (gui_info->process_to_gui, &ctrlmsg) == 0) {
    switch (ctrlmsg.type) {
      case CTRLMSG_ADD:
        chain_link = ctrlmsg.pointer;
        plugin = ctrlmsg.second_pointer;
        add_plugin_slot (gui_info, chain_link, plugin);
        break;
      case CTRLMSG_REMOVE:
        chain_link = ctrlmsg.pointer;
        plugin_slot = ctrlmsg.second_pointer;
        remove_plugin_slot (gui_info, plugin_slot);
        destroy_chain_link (chain_link);
        break;
      case CTRLMSG_MOVE:
        plugin_slot = ctrlmsg.second_pointer;
        move_plugin_slot (gui_info, plugin_slot,
                          GPOINTER_TO_INT (ctrlmsg.pointer));
        gui_info->moving_slot = FALSE;
        break;
      case CTRLMSG_CHANGE:
        change_plugin_slot (ctrlmsg.number, ctrlmsg.second_pointer, gui_info);
        destroy_chain_link (ctrlmsg.pointer);
        break;
      case CTRLMSG_ABLE:
        plugin_slot = ctrlmsg.second_pointer;
        enabled = GPOINTER_TO_INT(ctrlmsg.pointer);
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(plugin_slot->enable),
                                      enabled);
        plugin_slot->enabled = enabled;
      case CTRLMSG_QUIT: /* we'll never get a QUIT */
        break;
    }
  }
  
#ifdef HAVE_LADCCA
  if (cca_enabled (cca_client)) {
    cca_event_t * event;
    
    while ( (event = cca_get_event (cca_client)) ) {
      if (cca_event_get_type (event) == CCA_Quit) {
        quit_callback (NULL, gui_info);
      } else {
        printf ("recieved unknown LADCCA event of type %d\n", cca_event_get_type (event));
      }
      
      cca_event_destroy (event);
    }
  }
#endif
  
  return TRUE;
}

void jack_shutdown_callback (void * data) {
  guiinfo_t * gui_info;
  
  gui_info = data;
  
  gui_info->shutdown = TRUE;
}

